#include "Eigen/Dense"
#include <iostream>

template<typename T>
T f(const T &p, const T &x) { return p*x; }

template<typename T>
T dfdp(const T&, const T &x) { return x; }

template<typename T>
T ddfdpp(const T&, const T&) { return 0; }

template<typename T, int M>
T E(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T r=0;
  for (auto i=0;i<M;i++) r+=pow(f(p,x(i))-y(i),2);
  return r;
}

template<typename T, int M>
T dEdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T r=0;
  for (auto i=0;i<M;i++) r+=(f(p,x(i))-y(i))*dfdp(p,x(i));
  return 2*r;
}

template<typename T, int M>
T ddEdpp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T r=0;
  for (auto i=0;i<M;i++)
    r+=dfdp(p,x(i))*dfdp(p,x(i))+(f(p,x(i))-y(i))*ddfdpp(p,x(i));
  return 2*r;
}

template<typename T, int M>
void Newton(T& p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) {
  while (fabs(dEdp(p,x,y))>eps) {
    std::cout << ".";
    p-=dEdp(p,x,y)/ddEdpp(p,x,y);
  }
  std::cout << std::endl;
}

int main() {
  const int M=3;
  using T=double;
  using VT=Eigen::Matrix<T,M,1>;
  VT x=VT::Random(); x=x.cwiseProduct(x);
  std::cout << "x=^T" << x.transpose() << std::endl;
  VT y=VT::Random(); 
  std::cout << "y^T=" << y.transpose() << std::endl;
  T p=1;
  std::cout << "p=" << p << "; E=" << E(p,x,y) << "; dEdp=" << dEdp(p,x,y) 
	    << "; ddEdpp=" << ddEdpp(p,x,y) << std::endl;
  Newton(p,x,y,1e-7);
  std::cout << "p=" << p << "; E=" << E(p,x,y) << "; dEdp=" << dEdp(p,x,y) 
	    << "; ddEdpp=" << ddEdpp(p,x,y) << std::endl;
  return 0;
}
