#include "Eigen/Dense"
#include <iostream>
#include <cmath>
#include <limits>

#include "dco.hpp"

template<typename AT, typename PT>
AT f(const AT &p, const PT &x) { return pow(p*x,2); }

template<typename AT, typename PT, int M>
AT E(const AT &p, const Eigen::Matrix<PT,M,1> &x, const Eigen::Matrix<PT,M,1> &y) {
  AT o=0;
  for (int i=0;i<M;i++) o+=pow(f(p,x(i))-y(i),2);
  return o;
}

template<typename AT, typename PT, int M>
AT dEdp(const AT &p, const Eigen::Matrix<PT,M,1> &x, const Eigen::Matrix<PT,M,1> &y) {
  typename dco::gt1s<AT>::type p_t,r_t;
  dco::value(p_t)=p; dco::derivative(p_t)=1;
  r_t=E(p_t,x,y);
  return dco::derivative(r_t);
}

template<typename AT, typename PT, int M>
AT ddEdpp(const AT &p, const Eigen::Matrix<PT,M,1> &x, const Eigen::Matrix<PT,M,1> &y) {
  typename dco::gt1s<AT>::type p_t,r_t;
  dco::value(p_t)=p; dco::derivative(p_t)=1;
  r_t=dEdp(p_t,x,y);
  return dco::derivative(r_t);
}

template<typename T, int M>
void Newton(T& p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) {
  while (fabs(dEdp(p,x,y))>eps) {
    std::cout << ".";
    p-=dEdp(p,x,y)/ddEdpp(p,x,y);
  }
  std::cout << std::endl;
}

int main() {
  const int M=3;
  using T=double;
  using VT=Eigen::Matrix<T,M,1>;
  VT x=VT::Random(); x=x.cwiseProduct(x);
  std::cout << "x=^T" << x.transpose() << std::endl;
  VT y=VT::Random(); 
  std::cout << "y^T=" << y.transpose() << std::endl;
  T p=1;
  std::cout << "p=" << p << "; E=" << E(p,x,y) << "; dEdp=" << dEdp(p,x,y) 
	    << "; ddEdpp=" << ddEdpp(p,x,y) << std::endl;
  Newton(p,x,y,1e-7);
  std::cout << "p=" << p << "; E=" << E(p,x,y) << "; dEdp=" << dEdp(p,x,y) 
	    << "; ddEdpp=" << ddEdpp(p,x,y) << std::endl;
  return 0;
}

