#include "Eigen/Dense"
#include <iostream>
#include <cmath>
#include <limits>

template<typename T>
T f(const T &p, const T &x) { return pow(p*x,2); }

template<typename T, int M>
T E(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T o=0;
  for (int i=0;i<M;i++) o+=pow(f(p,x(i))-y(i),2);
  return o;
}

template<typename T, int M>
T dEdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T dp= (p==0) ? sqrt(sqrt(std::numeric_limits<T>::epsilon())) 
	       : sqrt(sqrt(std::numeric_limits<T>::epsilon()))*fabs(p);
  return (E(p+dp,x,y)-E(p-dp,x,y))/(2*dp);
}

template<typename T, int M>
T ddEdpp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T dp= (p==0) ? sqrt(sqrt(std::numeric_limits<T>::epsilon())) 
	       : sqrt(sqrt(std::numeric_limits<T>::epsilon()))*fabs(p);
  return (dEdp(p+dp,x,y)-dEdp(p-dp,x,y))/(2*dp);
}

template<typename T, int M>
void Newton(T& p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) {
  while (fabs(dEdp(p,x,y))>eps) {
    std::cout << ".";
    p-=dEdp(p,x,y)/ddEdpp(p,x,y);
  }
  std::cout << std::endl;
}

int main() {
  const int M=3;
  using T=double;
  using VT=Eigen::Matrix<T,M,1>;
  VT x=VT::Random(); x=x.cwiseProduct(x);
  std::cout << "x=^T" << x.transpose() << std::endl;
  VT y=VT::Random(); 
  std::cout << "y^T=" << y.transpose() << std::endl;
  T p=1;
  std::cout << "p=" << p << "; E=" << E(p,x,y) << "; dEdp=" << dEdp(p,x,y) 
	    << "; ddEdpp=" << ddEdpp(p,x,y) << std::endl;
  Newton(p,x,y,1e-7);
  std::cout << "p=" << p << "; E=" << E(p,x,y) << "; dEdp=" << dEdp(p,x,y) 
	    << "; ddEdpp=" << ddEdpp(p,x,y) << std::endl;
  return 0;
}

